import {parse as urlParse} from 'url';
import ExtendableError from 'es6-error';
export class RequestError extends ExtendableError {
  constructor(code, message, url, options = {}) {
    super(message);
    const {method, headers, body} = options;
    const parsed = urlParse(url);
    parsed.port = Number(parsed.port);
    Object.defineProperties(this, {
      code: {
        value: code
      },
      url: {
        value: parsed
      },
      path: {
        value: parsed.pathname
      },
      body: {
        value: body
      },
      method: {
        value: method
      },
      headers: {
        value: new Headers(headers || {})
      }
    });
    Object.freeze(this.url);
    Object.freeze(this.data);
    Object.freeze(this.headers);
  }
}
export default class Request {
  constructor({endpoint = '/api'} = {}) {
    Object.defineProperties(this, {
      endpoint: {
        value: endpoint,
        writable: true,
        enumerable: true
      }
    });
  }
  fetch(path, options = {}) {
    if (typeof (path) !== 'string') {
      throw new TypeError(`The path '${path}' must be a string not '${typeof path}'`);
    }
    const url = `${location.origin}${this.endpoint}/${path}`;
    return new Promise(resolve => (fetch.polyfill) ? setTimeout(resolve, 0) : resolve())
      .then(() => fetch(url, options))
      .then(response => {
        if (!response.ok) {
          return response.json().then(json => {
            throw new RequestError(response.status, json.error, url, options);
          });
        }
        return response;
      });
  }
  send({method, path, data, headers = {}}) {
    if (typeof (method) !== 'string') {
      throw new TypeError(`The method '${method}' must be a string`);
    }
    if (!(headers instanceof Headers)) {
      headers = new Headers(headers);
    }
    const options = {
      method,
      headers
    };
    if (!headers.has('Accept')) {
      headers.set('Accept', 'application/json');
    }
    if (data !== undefined) {
      const methodArray = ['PUT', 'POST'];
      const finder = methodArray.find(allowMethod => {
        return allowMethod === method;
      });
      if (!finder) {
        throw new TypeError(`Cannot send data with a '${method}' request`);
      }
      if (!headers.has('Content-Type')) {
        data = JSON.stringify(data);
        headers.set('Content-Type', 'application/json');
      }
      options.body = data;
    }
    return this.fetch(path, options)
      .then(response => response.status === 204 ? null : response.json());
  }
  get(path, headers = {}) {
    return this.send({method: 'GET', path, headers});
  }
  put(path, data, headers = {}) {
    return this.send({method: 'PUT', path, data, headers});
  }
  post(path, data, headers = {}) {
    return this.send({method: 'POST', path, data, headers});
  }
  delete(path, headers = {}) {
    return this.send({method: 'DELETE', path, headers});
  }
  options(path, headers = {}) {
    return this.send({method: 'OPTIONS', path, headers});
  }
  set(path, data) {
    return this.put(path, data);
  }
  openApi(path) {
    return this.options(path, {
      Accept: 'application/openapi+json'
    });
  }
  schema(path) {
    return this.options(path, {
      Accept: 'application/schema+json'
    });
  }
  add(path, data = {}) {
    return this.post(path, data).then(response => {
      if (!Object.isObject(response)) {
        throw new TypeError(`Adding an object at '${path}' was not an object: '${JSON.stringify(response)}'`);
      }
      if (!response.hasOwnProperty('index')) {
        throw new TypeError(`The response from the POST did not have an 'index' property: '${JSON.stringify(response)}'`);
      }
      if (!Number.isInteger(response.index)) {
        throw new TypeError(`Adding an object at '${path}' did not return a valid index value '${JSON.stringify(response.index)}' (${typeof response.index})`);
      }
      return response.index;
    });
  }
  erase(path) {
    return this.delete(path);
  }
}

//# sourceMappingURL=Request.js.map
