import Node from '../Node';
import FixedIdentifiable from '../identifiable/Fixed';
import ValueIdentifiable from '../identifiable/Value';
export default class Dynamic extends Node {
  constructor({schema, childSchema, emitter, endpoint}) {
    super({schema, endpoint, emitter, Data: Array});
    if (!Object.isObject(childSchema)) {
      throw new TypeError(`The schema for a 'Dynamic' node must have a valid child schema`);
    }
    Object.defineProperties(this, {
      _childSchema: {
        value: childSchema
      },
      _childClass: {
        value: childSchema.type === 'object' ? FixedIdentifiable : ValueIdentifiable
      }
    });
    Object.freeze(this._childSchema);
  }
  assign(object) {
    if (Array.isArray(object)) {
      for (let i = 0, length = this.length; i < length; ++i) {
        this[i].assign(object[i]);
      }
      for (let i = this.length, length = object.length; i < length; ++i) {
        this.push(object[i]);
      }
      this.length = object.length;
    } else {
      throw new TypeError(`Cannot assign a '${typeof object}' to a Dynamic node`);
    }
    return this;
  }
  createChild({data, id, dependency}) {
    return new this._childClass({
      data,
      id,
      dependency,
      schema: this._childSchema,
      emitter: this._emitter,
      endpoint: this._endpoint
    });
  }
  getById(id) {
    const value = this._data.find(identifiable => (identifiable._id === id));
    return (value instanceof ValueIdentifiable) ? value._data : value;
  }
  getByIndex(index) {
    return this._data[index];
  }
  setById(id, value) {
    this._data.find(identifiable => (identifiable._id === id)).assign(value);
  }
  setByIndex(index, value) {
    this._data[index].assign(value);
  }
  updateProperties(oldLength) {
    for (let i = 0, length = this.length; i < length; ++i) {
      if (!this.hasOwnProperty(i)) {
        Object.defineProperty(this, i, {
          get: this.getByIndex.bind(this, i),
          set: this.setByIndex.bind(this, i),
          enumerable: true,
          configurable: true
        });
      }
    }
    for (let i = this.length, old = oldLength || i; i < old; ++i) {
      delete this[i];
    }
  }
  get length() {
    return this._data.length;
  }
  set length(value) {
    if (this.length < value) {
      throw new TypeError(`Cannot extend the size of the local array to ${value}. Add new children with 'push', etc`);
    }
    this.splice(value);
  }
  every(callback, thisArg) {
    return this._data.every(callback, thisArg);
  }
  fill(value, start, end) {
    start = start || 0;
    end = end || this.length;
    this._data.slice(start, end).forEach(child => child.assign(value));
  }
  filter(callback, thisArg) {
    return this._data.filter(callback, thisArg);
  }
  find(callback, thisArg) {
    return this._data.find(callback, thisArg);
  }
  findIndex(callback, thisArg) {
    return this._data.findIndex(callback, thisArg);
  }
  forEach(callback, thisArg) {
    this._data.forEach(callback, thisArg);
  }
  includes(needle, from) {
    return this._data.includes(needle, from);
  }
  indexOf(needle, from) {
    return this._data.indexOf(needle, from);
  }
  lastIndexOf(needle, from) {
    return this._data.lastIndexOf(needle, from);
  }
  map(callback, thisArg) {
    return this._data.map(callback, thisArg);
  }
  pop() {
    return this.splice(this.length - 1, 1)[0];
  }
  push(...elements) {
    this.splice(this.length, 0, ...elements);
    return this.length;
  }
  reverse() {
    this._data.reverse();
  }
  shift() {
    return this.splice(0, 1)[0];
  }
  slice(begin, end) {
    return this._data.slice(begin, end);
  }
  some(callback, thisArg) {
    return this._data.every(callback, thisArg);
  }
  sort(compare) {
    this._data.sort(compare);
    return this;
  }
  splice(start, deleteCount, ...elements) {  
    throw new ReferenceError(`The 'splice' function must be implemented in the base class`);
  }
  toLocaleString() {
    return this._data.toLocaleString();
  }
  toString() {
    return this._data.toString();
  }
  unshift(...elements) {
    this.splice(0, 0, ...elements);
    return this.length;
  }
}

//# sourceMappingURL=Dynamic.js.map
