import Node from '../Node';
import Sparse from './Sparse';
import Vector from './Vector';
import Factory from './Factory';
import defaults from 'json-schema-defaults';
export default class Fixed extends Node {
  constructor({schema, emitter, endpoint, data}) {
    super({schema, endpoint, emitter});
    if (['object', 'array'].indexOf(this._schema.type) === -1) {
      throw new TypeError(`Schema must be object or array: ${this._schema.type}`);
    }
    this.defineProperties({data});
  }
  defineProperties({data, schema}) {
    schema = schema || this._schema;
    data = data || {};
    const properties = schema.properties;
    if (!Object.isObject(properties)) {
      throw new TypeError(`The schema for a 'Fixed' node must have a 'properties' key`);
    }
    for (const key in properties) {
      if (properties.hasOwnProperty(key)) {
        this.defineProperty(key, properties[key], data[key]);
      }
    }
  }
  defineProperty(key, schema, data) {
    if (schema.hasOwnProperty('patternProperties')) {
      const pp = schema.patternProperties;
      if (pp) {
        const keys = Object.keys(pp);
        const childKey = keys[0];
        const childSchema = pp[childKey];
        const Class = (childSchema.hasOwnProperty('oneOf')) ? Factory : Sparse;
        this._data[key] = new Class({
          data,
          schema,
          emitter: this._emitter,
          endpoint: `${this._endpoint}/${key}`
        });
      }
    } else if (schema.type === 'object') {
      this._data[key] = new Fixed({
        data,
        schema,
        emitter: this._emitter,
        endpoint: `${this._endpoint}/${key}`
      });
    } else if (schema.type === 'array') {
      this._data[key] = new Vector({
        data,
        schema,
        emitter: this._emitter,
        endpoint: `${this._endpoint}/${key}`
      });
    } else {
      const value = this._data[key] = data || defaults(schema);
      const result = this._validator.validate(value, schema);
      if (result.errors.length) {
        throw result.errors[0];
      }
    }
    Object.defineProperty(this, key, {
      get: this.getProperty.bind(this, key),
      set: this.setProperty.bind(this, key, schema),
      enumerable: true
    });
  }
  getProperty(key) {
    return this._data[key];
  }
  setProperty(key, schema, value) {
    const result = this._validator.validate(value, schema);
    if (result.errors.length) {
      throw result.errors[0];
    }
    if ((schema.type === 'object') || (schema.type === 'array')) {
      this._data[key].assign(value);
    } else if (this._data[key] !== value) {
      const old = this._data[key];
      this._data[key] = value;
      this.emit('changed', this, key, value, old);
      this._request.set(key, value)
        .then(response => {
          this.emit('committed', this, key, value, old, response);
        })
        .catch(e => {
          this.emit('revert', e, this, key, value, old);
        });
    }
  }
}

//# sourceMappingURL=Fixed.js.map
