import Dynamic from './Dynamic';
import FixedIdentifiable from '../identifiable/Fixed';
export default class Sparse extends Dynamic {
  constructor({schema, emitter, endpoint, data}) {
    const pp = schema.patternProperties;
    if (!Object.isObject(pp)) {
      throw new TypeError(`The schema for a 'Sparse' node must have a 'patternProperties' key`);
    }
    const keys = Object.keys(pp);
    if (keys.length !== 1) {
      throw new TypeError(`The schema for a 'Sparse' node must have only one pattern property`);
    }
    const regex = keys[0];
    const expected = '^[0-9]+$';
    if (regex !== expected) {
      throw new TypeError(`The pattern property '${regex}' for a 'Sparse' node must equal '${expected}'`);
    }
    const childSchema = pp[regex];
    super({schema, childSchema, endpoint, emitter});
    Object.defineProperty(this, '_regex', {
      value: new RegExp(regex)
    });
    data = data || {};
    for (const key in data) {
      if (data.hasOwnProperty(key)) {
        const id = Number(key);
        if (isNaN(id)) {
          throw new TypeError(`The inital data to the sparse node did not have a correct ID: ${key}`);
        }
        this._data.push(this.createChild({id, data: data[key]}));
      }
    }
    this.updateProperties();
  }
  toJSON() {
    const sparse = {};
    this._data
      .filter(identifiable => (identifiable._id !== null))
      .forEach(identifiable => {
        sparse[identifiable._id] = identifiable;
      });
    return sparse;
  }
  assign(object) {
    if (Object.isObject(object)) {
      for (const key in object) {
        if (!this._regex.exec(key)) {
          throw new TypeError(`Invalid property '${key}' assignment to a 'Sparse' node`);
        }
        if (this.hasOwnProperty(key)) {
          this[key] = object[key];
        } else {
          const id = Number(key);
          const data = object[key];
          const child = this.createChild({id, data});
          this.push(child);
        }
      }
      for (const key of Object.keys(this)) {
        if (!object.hasOwnProperty(key)) {
          const index = Number(key);
          this.splice(index, 1);
        }
      }
      return this;
    }
    return super.assign(object);
  }
  splice(start, deleteCount, ...elements) {
    const old = this.length;
    const added = elements.map(o => (o instanceof FixedIdentifiable) ? o : this.createChild({data: o}));
    deleteCount = (deleteCount === undefined) ? (old - start) : deleteCount;
    const removed = this._data.splice(start, deleteCount, ...added);
    if (added) {
      this.emit('added', this, start, added);
    }
    if (removed) {
      this.emit('erased', this, start, removed);
    }
    for (const child of removed) {
      const index = start++;
      child.loadId()
        .then(() => {
          return child._request.delete('');
        })
        .then(response => {
          this.emit('gone', child, this, index, response);
        }, e => {
          this.emit('reinsert', e, this, index, child);
        });
    }
    this.updateProperties(old);
    return removed;
  }
}

//# sourceMappingURL=Sparse.js.map
