import del from 'del';
import gulp from 'gulp';
import express from 'express';
import {Api} from 'vca-api-middleware';
import loadPlugins from 'gulp-load-plugins';
import babelify from 'babelify';
import browserify from 'browserify';
import vinylBuffer from 'vinyl-buffer';
import vinylSourceStream from 'vinyl-source-stream';
import {Server} from 'karma';
import {Instrumenter} from 'isparta';
import * as release from 'gitlab-release';

// Load all of our Gulp plugins
const $ = loadPlugins();

// Builds up the distribution of the files
gulp.task('dist-es5', ['clean'], () => {
  const bundler = browserify({
    entries: 'lib/vca-api.js',
    standalone: 'vca',
    debug: true
  });

  bundler.transform(babelify);

  return bundler.bundle()
    .pipe(vinylSourceStream('vca-api.js'))
    .pipe(vinylBuffer())
    .pipe($.sourcemaps.init({loadMaps: true}))
    .pipe($.uglify())
    .pipe($.rename({suffix: '.min'}))
    .pipe($.sourcemaps.write('.'))
    .pipe($.size({showFiles: true, showTotal: false}))
    .pipe(gulp.dest('dist'));
});

gulp.task('dist-es6', ['clean'], () => {
  return gulp.src(['lib/**/*.js'])
    .pipe($.sourcemaps.init({loadMaps: true}))
    .pipe($.stripComments())
    .pipe($.removeEmptyLines())
    .pipe($.sourcemaps.write('.'))
    .pipe(gulp.dest('dist'));
});

gulp.task('dist', ['dist-es5', 'dist-es6']);

// Cleans up the built files
gulp.task('clean', () => {
  return del('dist');
});

// Lints the source files
gulp.task('eslint', () => {
  return gulp.src(['gulpfile.babel.js', 'lib/**/*.js', 'test/**/*.js'])
    .pipe($.eslint())
    .pipe($.eslint.format())
    .pipe($.eslint.failAfterError());
});

// Performs static analysis
gulp.task('static', ['eslint']);

// Watches files for changes and reruns the static code analysis
gulp.task('watch-static', () => {
  gulp.watch(['gulpfile.babel.js', 'lib/**/*.js', 'test/**/*.js'], ['static']);
});

// Performs tests in the installed browsers
gulp.task('karma', done => {
  new Server({
    configFile: `${__dirname}/karma.conf.js`,
    singleRun: true
  })
  .on('run_complete', (browser, results) => {
    done(results.error ? 'There are test failures' : null);
  })
  .start();
});
for (const browser of ['Chrome', 'Firefox', 'PhantomJS', 'Safari', 'IE']) {
  gulp.task(`karma-${browser.toLowerCase()}`, done => {
    new Server({
      configFile: `${__dirname}/karma.conf.js`,
      singleRun: true,
      browsers: [browser],
      detectBrowsers: {
        enabled: false
      }
    })
    .on('run_complete', (browser, results) => {
      done(results.error ? 'There are test failures' : null);
    })
    .start();
  });
}

// Open browsers and re-runs tests whenever files change
gulp.task('watch-karma', done => {
  const server = new Server({
    configFile: `${__dirname}/karma.conf.js`
  }, done);
  server.start();
});

gulp.task('pre-mocha', () => {
  return gulp.src('lib/**/*.js')
    .pipe($.excludeGitignore())
    .pipe($.istanbul({
      includeUntested: true,
      instrumenter: Instrumenter
    }))
    .pipe($.istanbul.hookRequire());
});

gulp.task('mocha', ['pre-mocha'], cb => {
  let mochaErr;

  gulp.src(['test/mocha.js', 'test/spec/**/*.js'])
    .pipe($.plumber())
    .pipe($.mocha({
      reporter: 'spec'
    }))
    .on('error', err => {
      mochaErr = err;
      if (mochaErr.showStack) {
        cb(mochaErr);
      }
    })
    .pipe($.istanbul.writeReports())
    .on('end', () => {
      cb(mochaErr);
    });
});

gulp.task('watch-mocha', () => {
  gulp.watch(['lib/**/*.js', 'test/**/*.js'], ['mocha']);
});

// Performs tests in the installed browsers
gulp.task('test-sequence', ['mocha'], () => {
  return gulp.start('karma');
});
gulp.task('test', ['test-sequence']);

// Installs bower components
gulp.task('bower', () => $.bower({cmd: 'update'}));

// Runs the demo
gulp.task('demo', ['dist', 'bower'], () => {
  const api = new Api();
  const app = express();
  app.use(api.middleware.bind(api));
  app.get('/', (request, response) => {
    response.sendFile(`${__dirname}/demo.html`);
  });
  app.use(express.static(__dirname));
  const port = 8000;
  app.listen(port, () => {
    console.log(`Demo available at http://localhost:${port}`);
  });
});

// Release tasks
const releaseDataJson = '.release-data.json';

gulp.task('preversion', ['dist'], () => {
  const argv = JSON.parse(process.env.npm_config_argv);
  const type = argv.remain[0];
  return release.preversion({type, project: 'vca-api.js'})
    .then(data => release.writeJson(releaseDataJson, data))
    .then(() => release.gitAdd(['dist/**/*', 'dist/**/.*']));
});

gulp.task('postversion', () => {
  return release.readJson(releaseDataJson)
    .then(data => release.postversion({notes: data.description, project: 'vca-api.js'}))
    .then(() => release.gitRemove(['dist/**/*', 'dist/**/.*']))
    .then(() => release.gitCommit('Remove dist folder'))
    .then(() => release.gitCurrentBranch())
    .then(branch => release.gitPush({branches: [branch]}))
    .then(() => release.unlink(releaseDataJson));
});

// Ran before npm publishes the package
gulp.task('prepublish', ['dist']);

// An alias of test
gulp.task('default-sequence', ['static', 'dist'], () => {
  return gulp.start('test');
});
gulp.task('default', ['default-sequence']);
