import './api/identifiable/Fixed';
import Fixed from './api/node/Fixed';
import Endpoint from './api/Endpoint';

export default class Api extends Endpoint {
  constructor({endpoint = '/api', emitter} = {}) {
    super({endpoint, emitter});
    this.loadData();
  }

  /**
   * Initalizes the `data` property on the class. The schema and data is retrieved from the API and a node tree
   * respresentation is built from the data. Once the data is loaded the `data` callback is invoked.
   * @example
   * // Can use an event
   * const api = new Api();
   * api.on('data', data => {
   *   console.log('api.data is now available', data);
   * });
   *
   * // Using a promise
   * const api = new Api();
   * api.loadData().then(data => {
   *   console.log('api.data is now available', data);
   * });
   * @returns {Promise} resolves with the resolved data
   */
  loadData() {
    if (this.dataPromise) {
      return this.dataPromise;
    } else if (this.data) {
      return Promise.resolve(this.data);
    }

    this.dataPromise = Promise.all([this._request.get(''), this._request.schema('')])
      .then(([data, schema]) => {
        this.data = new Fixed({
          data,
          schema,
          emitter: this._emitter,
          endpoint: this._endpoint
        });

        /**
         * Raised when the data for the API has been fully loaded
         *
         * @event Api#data
         * @param {Object} data the data that was loaded
         * @param {Api} api the `Api` class instance that raised the event
         */
        this.emit('data', this.data, this);

        delete this.dataPromise;

        return this.data;
      });

    return this.dataPromise;
  }

  /**
   * Resets the API data to default values
   * @returns {Promise} resolves with the reset data
   */
  reset() {
    return this._request.post('reset')
      .then(() => {
        this.data.reset();
        return this.data;
      });
  }
}
