import EventEmitter from 'eventemitter2';

export default class Emitter {
  /**
   * A class that provides a subscriber/listener event system
   *
   * @param {EventEmitter} emitter the shared emitter
   * @public
   */
  constructor({emitter = new EventEmitter()}) {
    if ((typeof emitter !== 'object') || !(emitter instanceof EventEmitter)) {
      throw new TypeError(`Emitter must be EventEmitter`);
    }
    Object.defineProperty(this, '_emitter', {
      value: emitter
    });
  }

  /**
   * Emits an event to any listeners on the emitter. All events have the node as the first argument.
   * @param {String} name the event name to fire
   * @param {Iterable} args the arguments to invoke the event callbacks with
   * @example
   * this.emit('resolved', this, id);
   * @private
   */
  emit(name, ...args) {
    this._emitter.emit(name, ...args);
  }

  /**
   * Listens for an event on the node
   * @param {String} name the event name to fire
   * @param {Function} callback the function to invoke. The arguments passed with depend on the event
   * @example
   * this.emit('resolved', (node, id) => {
   *   console.log(node, id);
   * });
   * @private
   */
  on(name, callback) {
    this._emitter.on(name, callback);
  }

  /**
   * Adds a *one time* listener for an event
   * @param {String} name the event name to fire
   * @param {Function} callback the function to invoke. The arguments passed with depend on the event
   * @example
   * this.emit('resolved', (node, id) => {
   *   console.log(node, id);
   * });
   * @private
   */
  once(name, callback) {
    this._emitter.once(name, callback);
  }
}
