import {parse as urlParse} from 'url';
import ExtendableError from 'es6-error';

export class RequestError extends ExtendableError {
  constructor(code, message, url, options = {}) {
    super(message);
    const {method, headers, body} = options;
    const parsed = urlParse(url);
    parsed.port = Number(parsed.port);
    Object.defineProperties(this, {
      code: {
        value: code
      },
      url: {
        value: parsed
      },
      path: {
        value: parsed.pathname
      },
      body: {
        value: body
      },
      method: {
        value: method
      },
      headers: {
        value: new Headers(headers || {})
      }
    });
    Object.freeze(this.url);
    Object.freeze(this.data);
    Object.freeze(this.headers);
  }
}

export default class Request {
  constructor({endpoint = '/api'} = {}) {
    Object.defineProperties(this, {
      endpoint: {
        value: endpoint,
        writable: true,
        enumerable: true
      }
    });
  }

  /**
   * Performs an API fetch with the VCA APIs
   * @param {String} path the relative path from the end point
   * @param {Object} options the options that are passed through to the Fetch web API
   * @returns {Promise} resolves with the result of the fetch
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  fetch(path, options = {}) {
    if (typeof (path) !== 'string') {
      throw new TypeError(`The path '${path}' must be a string not '${typeof path}'`);
    }
    const url = `${location.origin}${this.endpoint}/${path}`;

    return new Promise(resolve => (fetch.polyfill) ? setTimeout(resolve, 0) : resolve())
      .then(() => fetch(url, options))
      .then(response => {
        if (!response.ok) {
          return response.json().then(json => {
            throw new RequestError(response.status, json.error, url, options);
          });
        }

        return response;
      });
  }

  /**
   * Performs an API request that returns the resulting JSON.
   * @param {String} method the HTTP method to perform
   * @param {String} path the relative path from the end point
   * @param {Object} [data] the JavaScript object to send as JSON to the API
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  send({method, path, data, headers = {}}) {
    if (typeof (method) !== 'string') {
      throw new TypeError(`The method '${method}' must be a string`);
    }

    if (!(headers instanceof Headers)) {
      headers = new Headers(headers);
    }

    const options = {
      method,
      headers
    };

    if (!headers.has('Accept')) {
      headers.set('Accept', 'application/json');
    }

    if (data !== undefined) {
      const methodArray = ['PUT', 'POST'];
      const finder = methodArray.find(allowMethod => {
        return allowMethod === method;
      });
      if (!finder) {
        throw new TypeError(`Cannot send data with a '${method}' request`);
      }
      if (!headers.has('Content-Type')) {
        data = JSON.stringify(data);
        headers.set('Content-Type', 'application/json');
      }
      options.body = data;
    }

    return this.fetch(path, options)
      .then(response => response.status === 204 ? null : response.json());
  }

  /**
   * Performs a `GET` HTTP request on the API
   * @param {String} path the relative path from the end point
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  get(path, headers = {}) {
    return this.send({method: 'GET', path, headers});
  }

  /**
   * Performs a `PUT` HTTP request on the API
   * @param {String} path the relative path from the end point
   * @param {Object} data the JavaScript object to send as JSON to the API
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  put(path, data, headers = {}) {
    return this.send({method: 'PUT', path, data, headers});
  }

  /**
   * Performs a `PUT` HTTP request on the API
   * @param {String} path the relative path from the end point
   * @param {Object} data the JavaScript object to send as JSON to the API
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  post(path, data, headers = {}) {
    return this.send({method: 'POST', path, data, headers});
  }

  /**
   * Performs a `DELETE` HTTP request on the API
   * @param {String} path the relative path from the end point
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  delete(path, headers = {}) {
    return this.send({method: 'DELETE', path, headers});
  }

  /**
   * Performs an `OPTIONS` HTTP request on the API
   * @param {String} path the relative path from the end point
   * @param {Object|Headers} [headers] the headers that should be set with the request
   * @returns {Promise} resolves with the result of the request
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   * @private
   */
  options(path, headers = {}) {
    return this.send({method: 'OPTIONS', path, headers});
  }

  /**
   * Sets the value(s) of the configuration
   * @param {String} path the relative path from the end point
   * @param {Object} data the JavaScript object to send as JSON to the API
   * @returns {Promise} resolves when the data has been set
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   */
  set(path, data) {
    return this.put(path, data);
  }

  /**
   * Retrieves the OpenAPI representation of the API
   * @param {String} path the relative path from the end point
   * @returns {Promise} resolves when the data has been set
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   */
  openApi(path) {
    return this.options(path, {
      Accept: 'application/openapi+json'
    });
  }

  /**
   * Retrieves the schema representation of the API
   * @param {String} path the relative path from the end point
   * @returns {Promise} resolves when the data has been set
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   */
  schema(path) {
    return this.options(path, {
      Accept: 'application/schema+json'
    });
  }

  /**
   * Adds a new value into the configuration
   * @param {String} path the relative path from the end point
   * @param {Object} data the JavaScript object to send as JSON to the API
   * @returns {Promise} resolves with the index of the new object
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   */
  add(path, data = {}) {
    return this.post(path, data).then(response => {
      if (!Object.isObject(response)) {
        throw new TypeError(`Adding an object at '${path}' was not an object: '${JSON.stringify(response)}'`);
      }
      if (!response.hasOwnProperty('index')) {
        throw new TypeError(`The response from the POST did not have an 'index' property: '${JSON.stringify(response)}'`);
      }
      if (!Number.isInteger(response.index)) {
        throw new TypeError(`Adding an object at '${path}' did not return a valid index value '${JSON.stringify(response.index)}' (${typeof response.index})`);
      }
      return response.index;
    });
  }

  /**
   * Removes a previously added object in the configuration
   * @param {String} path the relative path from the end point
   * @returns {Promise} resolves when the object has been erased
   * @throws {TypeError} when there is a network error
   * @throws {RequestError} the API returned a bad status code
   */
  erase(path) {
    return this.delete(path);
  }
}
