import Node from '../Node';
import FixedIdentifiable from '../identifiable/Fixed';
import ValueIdentifiable from '../identifiable/Value';

export default class Dynamic extends Node {
  constructor({schema, childSchema, emitter, endpoint}) {
    super({schema, endpoint, emitter, Data: Array});

    if (!Object.isObject(childSchema)) {
      throw new TypeError(`The schema for a 'Dynamic' node must have a valid child schema`);
    }

    Object.defineProperties(this, {
      _childSchema: {
        value: childSchema
      },
      _childClass: {
        value: childSchema.type === 'object' ? FixedIdentifiable : ValueIdentifiable
      }
    });
    Object.freeze(this._childSchema);
  }

  /** @inheritdoc */
  assign(object) {
    if (Array.isArray(object)) {
      for (let i = 0, length = this.length; i < length; ++i) {
        this[i].assign(object[i]);
      }
      for (let i = this.length, length = object.length; i < length; ++i) {
        this.push(object[i]);
      }
      this.length = object.length;
    } else {
      throw new TypeError(`Cannot assign a '${typeof object}' to a Dynamic node`);
    }
    return this;
  }

  /**
   * Creates a new child to be added to the dynamic array
   * @param {Object} object data to be set on this object
   * @param {Number} id the identification number for the child. The API will not be requested to provide an ID.
   * @param {Identifiable} dependency another child that must resolve it's ID before this new child does
   * @returns {Identifiable} the new identifiable child
   * @private
   */
  createChild({data, id, dependency}) {
    return new this._childClass({
      data,
      id,
      dependency,
      schema: this._childSchema,
      emitter: this._emitter,
      endpoint: this._endpoint
    });
  }

  /**
   * Retrieves a child by it's identification number
   * @param {Number} id the index of the child in the API sparse array
   * @returns {Identifiable} the child matching the `id` or `undefined`
   */
  getById(id) {
    const value = this._data.find(identifiable => (identifiable._id === id));
    return (value instanceof ValueIdentifiable) ? value._data : value;
  }

  /**
   * Retrieves a child by it's index in the local `Array`
   * @param {Number} index the array index
   * @returns {Identifiable} the child at the index or `undefined`
   */
  getByIndex(index) {
    return this._data[index];
  }

  /**
   * Sets a child by it's identification number
   * @param {Number} id the index of the child in the API sparse array
   * @param {Object} value the object to assign to the child
   */
  setById(id, value) {
    this._data.find(identifiable => (identifiable._id === id)).assign(value);
  }

  /**
   * Sets a child by it's index in the local `Array`
   * @param {Number} index the array index
   * @param {Object} value the object to assign to the child
   */
  setByIndex(index, value) {
    this._data[index].assign(value);
  }

  /**
   * Makes sure the the index properties are correct for this object. Can be used when the underlying array is modified
   * to expose the correct properties on this object to mimic the underlying array.
   * @param {Number} [oldLength] the previous length so that properties can be removed
   * @private
   */
  updateProperties(oldLength) {
    for (let i = 0, length = this.length; i < length; ++i) {
      if (!this.hasOwnProperty(i)) {
        Object.defineProperty(this, i, {
          get: this.getByIndex.bind(this, i),
          set: this.setByIndex.bind(this, i),
          enumerable: true,
          configurable: true
        });
      }
    }
    for (let i = this.length, old = oldLength || i; i < old; ++i) {
      delete this[i];
    }
  }

  /**
   * @property {Number} property length the number of children this sparse node has
   */
  get length() {
    return this._data.length;
  }
  set length(value) {
    if (this.length < value) {
      throw new TypeError(`Cannot extend the size of the local array to ${value}. Add new children with 'push', etc`);
    }
    this.splice(value);
  }

  /**
   * Tests whether all elements in the array pass the test implemented by the provided function.
   * @param {Arguments} callback to test each element of the array. Invoked with arguments `(element, index, array)`.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Array} the filtered array
   */
  every(callback, thisArg) {
    return this._data.every(callback, thisArg);
  }

  /**
   * Fills all the children from a start index to an end index with a static value.
   * @param {Object} value the static value to set. Will be assigned to each child
   * @param {Number} [start=0] the start of the range to apply the `value` to
   * @param {Number} [end=this.length] the index to apply the `value` up to
   */
  fill(value, start, end) {
    start = start || 0;
    end = end || this.length;
    this._data.slice(start, end).forEach(child => child.assign(value));
  }

  /**
   * Creates a new array with all elements that pass the test implemented by the provided function.
   * @param {Arguments} callback to test each element of the array. Invoked with arguments `(element, index, array)`.
   * Return `true` to keep the element, `false` otherwise.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Array} the filtered array
   */
  filter(callback, thisArg) {
    return this._data.filter(callback, thisArg);
  }

  /**
   * Returns a value in the array, if an element in the array satisfies the provided testing function.
   * Otherwise undefined is returned.
   * @param {Arguments} callback to execute on each value in the array, taking three arguments
   * `(element, index, array)`. Return `true` to return the element, false otherwise.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Identifiable} the found value or `undefined`
   */
  find(callback, thisArg) {
    return this._data.find(callback, thisArg);
  }

  /**
   * Returns an index in the array, if an element in the array satisfies the provided testing function.
   * Otherwise undefined is returned.
   * @param {Arguments} callback to execute on each value in the array, taking three arguments
   * `(element, index, array)`. Return `true` to return the index, false otherwise.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Identifiable} the found index or `undefined`
   */
  findIndex(callback, thisArg) {
    return this._data.findIndex(callback, thisArg);
  }

  /**
   * Executes a provided function once per array element.
   * @param {Arguments} callback to execute on each value in the array, taking three arguments
   * `(element, index, array)`. Return `true` to return the index, false otherwise.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   */
  forEach(callback, thisArg) {
    this._data.forEach(callback, thisArg);
  }

  /**
   * Determines whether an array includes a certain element, returning `true` or `false` as appropriate.
   * @param {Identifiable} needle the element to search for
   * @param {Number} [from] The position in this array at which to begin searching for `needle`. A negative value
   * searches from the end of the array. Defaults to `0`.
   * @returns {Boolean} `true` or `false`
   */
  includes(needle, from) {
    return this._data.includes(needle, from);
  }

  /**
   * Returns the first index at which a given element can be found in the array, or `-1` if it is not present.
   * @param {Identifiable} needle the element to search for
   * @param {Number} [from] The position in this array at which to begin searching for `needle`. A negative value
   * searches from the end of the array. Defaults to `0`.
   * @returns {Boolean} `true` or `false`
   */
  indexOf(needle, from) {
    return this._data.indexOf(needle, from);
  }

  /**
   * Returns the last index at which a given element can be found in the array, or `-1` if it is not present.
   * @param {Identifiable} needle the element to search for
   * @param {Number} [from] The position in this array at which to begin searching for `needle`. A negative value
   * searches from the end of the array. Defaults to `0`.
   * @returns {Boolean} `true` or `false`
   */
  lastIndexOf(needle, from) {
    return this._data.lastIndexOf(needle, from);
  }

  /**
   * Creates a new array with the results of calling a provided function on every element in this array.
   * @param {Arguments} callback to execute on each value in the array, taking three arguments
   * `(element, index, array)`. Return the mapped value for the new array.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Array} the new mapped array
   */
  map(callback, thisArg) {
    return this._data.map(callback, thisArg);
  }

  /**
   * Removes the last element from the local array and returns it to the caller. The removed local element will be
   * erased in the web API.
   * @returns {Identifiable} the local node that will have been removed from the API
   */
  pop() {
    return this.splice(this.length - 1, 1)[0];
  }

  /**
   * Pushes new identifiable children to the end of the sparse array
   * @param {Arguments} elements the children to add to the array
   * @returns {Number} the new length of the local array
   */
  push(...elements) {
    this.splice(this.length, 0, ...elements);
    return this.length;
  }

  /**
   * Reverses the order of the local children
   */
  reverse() {
    this._data.reverse();
  }

  /**
   * Removes the first element from the local array and returns it to the caller. The removed local element will be
   * erased in the web API.
   * @returns {Identifiable} the local node that will have been removed from the API
   */
  shift() {
    return this.splice(0, 1)[0];
  }

  /**
   * Returns a shallow copy of a portion of an array into a new array object.
   * @param {Number} begin zero-based index at which to begin extraction. As a negative index, `begin` indicates an
   * offset from the end of the sequence. `slice(-2)` extracts the last two elements in the sequence. If begin is
   * `undefined`, slice begins from index `0`.
   * @param {Number} end zero-based index at which to end extraction. slice extracts up to but not including `end`.
   * `slice(1,4)` extracts the second element through the fourth element (elements indexed `1`, `2`, and `3`). As a
   * negative index, `end` indicates an offset from the end of the sequence. `slice(2,-1)` extracts the third element
   * through the second-to-last element in the sequence. If `end` is omitted, slice extracts through the end of the
   * sequence (`arr.length).
   * @returns {Array} the shallow copy of the local arrays
   */
  slice(begin, end) {
    return this._data.slice(begin, end);
  }

  /**
   * Tests whether some of elements in the array pass the test implemented by the provided function.
   * @param {Arguments} callback to test each element of the array. Invoked with arguments `(element, index, array)`.
   * @param {Object} [thisArg] Value to use as `this` when executing `callback`.
   * @returns {Array} the filtered array
   */
  some(callback, thisArg) {
    return this._data.every(callback, thisArg);
  }

  /**
   * Sorts the elements of an array in place
   * @param {Function} [compare] a function that defines the sort order. If omitted, the array is sorted according to
   * each character's Unicode code point value, according to the string conversion of each element.
   * @returns {Dynamic} this node
   */
  sort(compare) {
    this._data.sort(compare);
    return this;
  }

  // eslint-disable valid-jsdoc
  /**
   * Changes the content of the array by removing existing elements and/or adding new elements
   * @param {Number} start index at which to start changing the array (with origin `0`). If greater than the length of
   * the array, actual starting index will be set to the length of the array. If negative, will begin that many elements
   * from the end.
   * @param {Number} deleteCount an integer indicating the number of old array elements to remove. If `deleteCount` is
   * `0`, no elements are removed. In this case, you should specify at least one new element. If `deleteCount` is
   * greater than the number of elements left in the array starting at start, then all of the elements through the end
   * of the array will be deleted. If `deleteCount` is omitted, `deleteCount` will be equal to `(arr.length - start)`.
   * @param {Arguments} elements the elements to add to the array, beginning at the `start` index. If you don't
   * specify any elements, `splice()` will only remove elements from the array.
   * @returns {Array} the removed elements
   */
  // eslint-enable valid-jsdoc
  splice(start, deleteCount, ...elements) {  // eslint-disable-line no-unused-vars
    throw new ReferenceError(`The 'splice' function must be implemented in the base class`);
  }

  /**
   * Rreturns a string representing the elements of the array. The elements are converted to `Strings` using their
   * `toLocaleString` methods and these `Strings` are separated by a locale-specific `String` (such as a comma )
   * @returns {String} the localized string
   */
  toLocaleString() {
    return this._data.toLocaleString();
  }

  /**
   * Returns a string representing the specified array and its elements.
   * @returns {String} the representation of the local children
   */
  toString() {
    return this._data.toString();
  }

  /**
   * Pushes new identifiable children to the start of the sparse array
   * @param {Arguments} elements the children to add to the array
   * @returns {Number} the new length of the local array
   */
  unshift(...elements) {
    this.splice(0, 0, ...elements);
    return this.length;
  }
}
