import Node from '../Node';
import Sparse from './Sparse';
import Vector from './Vector';
import Factory from './Factory';
import defaults from 'json-schema-defaults';

export default class Fixed extends Node {
  constructor({schema, emitter, endpoint, data}) {
    super({schema, endpoint, emitter});

    if (['object', 'array'].indexOf(this._schema.type) === -1) {
      throw new TypeError(`Schema must be object or array: ${this._schema.type}`);
    }

    this.defineProperties({data});
  }

  /**
   * Defines the properties from the schema on this node
   * @param {Object} [data] the intial data to use for the properties
   * @param {Object} [schema=this._schema] the schema to use to define the properties
   * @private
   */
  defineProperties({data, schema}) {
    schema = schema || this._schema;
    data = data || {};
    const properties = schema.properties;
    if (!Object.isObject(properties)) {
      throw new TypeError(`The schema for a 'Fixed' node must have a 'properties' key`);
    }
    for (const key in properties) {
      if (properties.hasOwnProperty(key)) {
        this.defineProperty(key, properties[key], data[key]);
      }
    }
  }

  /**
   * Defines a property on the `Node`. This will recursively create nodes if the schema is an `object` type.
   * @param {String} key the property name
   * @param {Object} schema the schema of the current property
   * @param {Object} [data] the intial value of the property
   * @private
   */
  defineProperty(key, schema, data) {
    if (schema.hasOwnProperty('patternProperties')) {
      const pp = schema.patternProperties;
      if (pp) {
        const keys = Object.keys(pp);
        const childKey = keys[0];
        const childSchema = pp[childKey];
        const Class = (childSchema.hasOwnProperty('oneOf')) ? Factory : Sparse;
        this._data[key] = new Class({
          data,
          schema,
          emitter: this._emitter,
          endpoint: `${this._endpoint}/${key}`
        });
      }
    } else if (schema.type === 'object') {
      this._data[key] = new Fixed({
        data,
        schema,
        emitter: this._emitter,
        endpoint: `${this._endpoint}/${key}`
      });
    } else if (schema.type === 'array') {
      this._data[key] = new Vector({
        data,
        schema,
        emitter: this._emitter,
        endpoint: `${this._endpoint}/${key}`
      });
    } else {
      const value = this._data[key] = data || defaults(schema);
      const result = this._validator.validate(value, schema);
      if (result.errors.length) {
        throw result.errors[0];
      }
    }
    Object.defineProperty(this, key, {
      get: this.getProperty.bind(this, key),
      set: this.setProperty.bind(this, key, schema),
      enumerable: true
    });
  }

  /**
   * Returns the value from the data in the property key
   * @param {String} key the property name
   * @returns {Promise} resolves with the value of the
   * @private
   */
  getProperty(key) {
    return this._data[key];
  }

  /**
   * Sets the value property. The new value will be validated against the schema and a `ValidationError` will be
   * thrown if it does not pass. Once the value has passed validation it will be set locally and will send off a
   * request to set the value in the API. If the API request fails then the `revert` callback will be invoked so that
   * the user can decide to revert the local value back to the old data.
   * @private
   * @param {String} key the property name
   * @param {Object} schema the schema of the current property
   * @param {Object} value the new value to be set
   * @throws {ValidationError} if the value does not match the expected value format
   * @see {@link Base} for callback `revert` documentation
  */
  setProperty(key, schema, value) {
    const result = this._validator.validate(value, schema);
    if (result.errors.length) {
      throw result.errors[0];
    }
    if ((schema.type === 'object') || (schema.type === 'array')) {
      this._data[key].assign(value);
    } else if (this._data[key] !== value) {
      const old = this._data[key];
      this._data[key] = value;

      /**
       * A property has changed on the API object
       *
       * @event Fixed#changed
       * @param {Node} node the node that has had a property change
       * @param {String} key the name of the property that has been modified
       * @param {Object} value the new value of the property
       * @param {Object} [old] the value of the property before the change occurred
       */
      this.emit('changed', this, key, value, old);

      this._request.set(key, value)
        .then(response => {
          /**
           * A changed property has been successfully committed to the API
           *
           * @event Fixed#committed
           * @param {Node} node the node that had the property change
           * @param {String} key the name of the property that has been modified
           * @param {Object} value the new value of the property
           * @param {Object} [old] the value of the property before the change occurred
           * @param {Object} [response] the response from the web API
           */
          this.emit('committed', this, key, value, old, response);
        })
        .catch(e => {
          /**
           * A changed property failed to be set in the web API. This event allows the user to revert the local copy
           *
           * @event Fixed#revert
           * @param {RequestError} e the error that occurred on the web API
           * @param {Node} node the node that had the property change
           * @param {String} key the name of the property that has been modified
           * @param {Object} value the new value of the property
           * @param {Object} [old] the value of the property before the change occurred
           */
          this.emit('revert', e, this, key, value, old);
        });
    }
  }
}
