import Dynamic from './Dynamic';
import FixedIdentifiable from '../identifiable/Fixed';

export default class Sparse extends Dynamic {
  constructor({schema, emitter, endpoint, data}) {
    const pp = schema.patternProperties;
    if (!Object.isObject(pp)) {
      throw new TypeError(`The schema for a 'Sparse' node must have a 'patternProperties' key`);
    }
    const keys = Object.keys(pp);
    if (keys.length !== 1) {
      throw new TypeError(`The schema for a 'Sparse' node must have only one pattern property`);
    }

    const regex = keys[0];
    const expected = '^[0-9]+$';
    if (regex !== expected) {
      throw new TypeError(`The pattern property '${regex}' for a 'Sparse' node must equal '${expected}'`);
    }
    const childSchema = pp[regex];

    super({schema, childSchema, endpoint, emitter});

    Object.defineProperty(this, '_regex', {
      value: new RegExp(regex)
    });

    data = data || {};
    for (const key in data) {
      if (data.hasOwnProperty(key)) {
        const id = Number(key);
        if (isNaN(id)) {
          throw new TypeError(`The inital data to the sparse node did not have a correct ID: ${key}`);
        }
        this._data.push(this.createChild({id, data: data[key]}));
      }
    }
    this.updateProperties();
  }

  /**
   * Converts the object to JSON. Whilst the JavaScript object works as an `Array` the actually API object is a sparse
   * array. The JSON output reflects this.
   * @returns {Object} the sparse representation to serialize
   * @private
   */
  toJSON() {
    const sparse = {};
    this._data
      .filter(identifiable => (identifiable._id !== null))
      .forEach(identifiable => {
        sparse[identifiable._id] = identifiable;
      });
    return sparse;
  }

  /** @inheritdoc */
  assign(object) {
    if (Object.isObject(object)) {
      // Add new objects
      for (const key in object) {
        if (!this._regex.exec(key)) {
          throw new TypeError(`Invalid property '${key}' assignment to a 'Sparse' node`);
        }
        if (this.hasOwnProperty(key)) {
          this[key] = object[key];
        } else {
          const id = Number(key);
          const data = object[key];
          const child = this.createChild({id, data});
          this.push(child);
        }
      }

      // Remove old objects.
      for (const key of Object.keys(this)) {
        if (!object.hasOwnProperty(key)) {
          const index = Number(key);
          this.splice(index, 1);
        }
      }

      return this;
    }

    return super.assign(object);
  }

  /** @inheritdoc */
  splice(start, deleteCount, ...elements) {
    // Store the length so we can update the properties
    const old = this.length;

    // Convert the children to `Identifiable`s
    const added = elements.map(o => (o instanceof FixedIdentifiable) ? o : this.createChild({data: o}));

    // May need to remove up to the end of the array
    deleteCount = (deleteCount === undefined) ? (old - start) : deleteCount;

    // Do the local splice
    const removed = this._data.splice(start, deleteCount, ...added);

    // Raise the local events
    if (added) {
      /**
       * Raised when a new dynamic children are added
       *
       * @event Sparse#resolved
       * @param {Sparse} node the node that has new children
       * @param {Number} start the index to the start of the new children
       * @param {Number} children the array of the newly added children
       */
      this.emit('added', this, start, added);
    }
    if (removed) {
      /**
       * Raised when a dynamic children are erased
       *
       * @event Sparse#erased
       * @param {Sparse} node the node that has been modified
       * @param {Number} start the index where the children were removed from
       * @param {Number} children the array of the erased children
       */
      this.emit('erased', this, start, removed);
    }

    // Remove the erased elements from the web API
    for (const child of removed) {
      const index = start++;
      child.loadId()
        .then(() => {
          // Child is now available in the web API so remove it
          return child._request.delete('');
        })
        .then(response => {
          /**
           * Raised when a dynamic child has been removed from the web API
           *
           * @event Sparse#gone
           * @param {Identifiable} child the removed child
           * @param {Sparse} node the node that the child was removed from
           * @param {Number} index the location in the local node array that the child was before removal
           * @param {Object} response the returned data from the web API
           */
          this.emit('gone', child, this, index, response);
        }, e => {
          /**
           * Raised when a dynamic node fails to be removed from the web API. Can be used to reinsert the node back into
           * the local array.
           *
           * @event Sparse#gone
           * @param {RequestError} e the error that occurred in the web API
           * @param {Sparse} node the node that the child was removed from
           * @param {Number} index the location in the local node array that the child was before removal
           * @param {Identifiable} child the removed child
           */
          this.emit('reinsert', e, this, index, child);
        });
    }

    // Update our properties
    this.updateProperties(old);

    return removed;
  }
}
