import Dynamic from './Dynamic';
import defaults from 'json-schema-defaults';

export default class Vector extends Dynamic {
  constructor({schema, emitter, endpoint, data}) {
    const items = schema.items;
    if (!Object.isObject(items)) {
      throw new TypeError(`The schema for a 'Vector' node must have a 'items' key`);
    }

    super({schema, endpoint, emitter, childSchema: schema.items});

    const length = this.length;
    const children = data || defaults(this._schema);
    this._data.push(...children.map((data, id) => this.createChild({data, id})));
    this.updateProperties(length);
  }

  /**
   * Converts the object to JSON.
   * @returns {Object} the actual underlying array
   * @private
   */
  toJSON() {
    return this._data;
  }

  /** @inheritdoc */
  splice(start, deleteCount, ...elements) {
    // Store the length so we can update the properties
    const old = this.length;

    // May need to remove up to the end of the array
    deleteCount = (deleteCount === undefined) ? (old - start) : deleteCount;

    // We want the local IDs to stay the same as the API, so we leave the local children in place
    const copy = this._data.slice();

    // Do the local splice
    const removed = copy.splice(start, deleteCount, ...elements).map(o => this.createChild({data: o, id: o._id}));

    // Add any new children needed to the API
    let child = null;
    for (let i = this.length, length = copy.length; i < length; ++i) {
      child = this.createChild({data: copy[i], dependency: child});
      this._data.push(child);
    }

    // Update any data in the children, which will update the API
    for (let i = 0, length = old; i < length; ++i) {
      this._data[i].assign(copy[i]);
    }

    // Determine the 'added' nodes
    const added = this._data.slice(old);

    // Remove any that actually need to be deleted in the API
    const gone = this._data.splice(copy.length);

    // Raise the local events
    if (added) {
      /**
       * Raised when a new dynamic children are added
       *
       * @event Sparse#resolved
       * @param {Sparse} node the node that has new children
       * @param {Number} start the index to the start of the new children
       * @param {Number} children the array of the newly added children
       */
      this.emit('added', this, start, added);
    }
    if (removed) {
      /**
       * Raised when a dynamic children are erased
       *
       * @event Sparse#erased
       * @param {Sparse} node the node that has been modified
       * @param {Number} start the index where the children were removed from
       * @param {Number} children the array of the erased children
       */
      this.emit('erased', this, start, removed);
    }

    // Remove the erased elements from the web API
    for (const child of gone) {
      child.loadId()
        .then(() => {
          // Child is now available in the web API so remove it
          return child._request.delete('');
        })
        .then(response => {
          /**
           * Raised when a dynamic child has been removed from the web API
           *
           * @event Sparse#gone
           * @param {Identifiable} child the removed child
           * @param {Sparse} node the node that the child was removed from
           * @param {Number} index the location in the local node array that the child was before removal
           * @param {Object} response the returned data from the web API
           */
          this.emit('gone', this, child._id, child, response);
        }, e => {
          /**
           * Raised when a dynamic node fails to be removed from the web API. Can be used to reinsert the node back into
           * the local array.
           *
           * @event Sparse#gone
           * @param {RequestError} e the error that occurred in the web API
           * @param {Sparse} node the node that the child was removed from
           * @param {Number} index the location in the local node array that the child was before removal
           * @param {Identifiable} child the removed child
           */
          this.emit('reinsert', e, this, child._id, child);
        });
    }

    // Update our properties
    this.updateProperties(old);

    return removed;
  }
}
