/* eslint max-nested-callbacks: 0 */

import Api from '../../lib/Api.js';
import {ValidationError} from '../../lib/api/Node';
import defaults from 'json-schema-defaults';

describe('Api', () => {
  beforeEach(function () {
    this.api = new Api();
    return this.api.loadData();
  });

  it('can be default constructed', function () {
    this.api.should.be.instanceOf(Api);
    const defaultVersion = defaults(this.api.data.software.vca.api.middleware.version._schema);
    this.api.data.software.vca.api.middleware.version.should.deep.equal(defaultVersion);
  });

  describe('data', () => {
    it('has the correct version structure', function () {
      this.api.data.should.have.deep.property('software.vca.api.middleware.version')
        .that.has.all.keys('major', 'minor', 'patch', 'commit');
    });

    it('should be able to change the major value', function (done) {
      const newValue = 12;
      this.api.on('committed', () => done());
      this.api.on('revert', done);
      this.api.data.software.vca.api.middleware.version.major = newValue;
      this.api.data.software.vca.api.middleware.version.major.should.equal(newValue);
    });

    it('should raise an exception for bad data', function () {
      const newValue = '20';
      const fun = () => {
        this.api.data.software.vca.api.middleware.version.major = newValue;
      };
      fun.should.throw(ValidationError);
    });

    describe('zones', () => {
      beforeEach(function () {
        this.api.data.zones.length.should.equal(0, 'the zones were not empty before the test started');
      });

      it('can push new zones', function () {
        const index = this.api.data.zones.push({}) - 1;
        const zone = this.api.data.zones[index];
        return zone.loadId()
          .then(id => {
            this.api.data.zones.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('sets the nested endpoints correctly', function () {
        const index = this.api.data.zones.push({}) - 1;
        const zone = this.api.data.zones[index];
        return zone.loadId()
          .then(id => {
            return this.api.data.zones[0].colour._endpoint.should.equal(`/api/zones/${id}/colour`);
          });
      });

      it('can add lots of zones', function () {
        this.api.data.zones.splice(0, 0, {}, {name: 'blah'}, {}, {detection: false});
        this.api.data.zones.length.should.equal(4);
        return Promise.all(this.api.data.zones.map(zone => zone.loadId()))
          .then(() => this.api._request.get('zones'))
          .then(zones => {
            Object.keys(zones).should.have.length(4);
          });
      });

      it('can unshift new zones', function () {
        this.api.data.zones.unshift({});
        return this.api.data.zones[0].loadId()
          .then(id => {
            this.api.data.zones.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can add named zones', function () {
        const name = 'Hello, world!';
        const index = this.api.data.zones.push({name}) - 1;
        const zone = this.api.data.zones[index];
        zone.name.should.equal(name);
        return zone.loadId()
          .then(() => {
            this.api.data.zones.length.should.equal(1);
            return this.api.data.zones[0]._request.get('');
          })
          .should.eventually.have.property('name')
          .that.equals(name);
      });

      it('can shorten length', function (done) {
        const index = this.api.data.zones.push({}) - 1;
        const zone = this.api.data.zones[index];
        zone.loadId()
          .then(() => {
            this.api.data.zones.length.should.equal(1);
            this.api.data.zones.on('reinsert', done);
            this.api.data.zones.on('gone', () => done());
            this.api.data.zones.length = 0;
          }).catch(done);
      });

      it('cannot extend length', function () {
        const func = () => {
          this.api.data.zones.length = 10;
        };
        func.should.throw(TypeError);
      });

      it('can pop zones', function (done) {
        const name = 'Hello, world!';
        const index = this.api.data.zones.push({name}) - 1;
        const zone = this.api.data.zones[index];
        zone.loadId()
          .then(() => {
            this.api.data.zones.length.should.equal(1);
            this.api.data.zones.on('reinsert', done);
            this.api.data.zones.on('gone', () => done());
            this.api.data.zones.pop().should.have.property('name').that.equals(name);
          }).catch(done);
      });

      it('can shift zones', function (done) {
        const name = 'Hello, world!';
        const index = this.api.data.zones.push({name}) - 1;
        const zone = this.api.data.zones[index];
        zone.loadId()
          .then(() => {
            this.api.data.zones.length.should.equal(1);
            this.api.data.zones.on('reinsert', done);
            this.api.data.zones.on('gone', () => done());
            this.api.data.zones.shift().should.have.property('name').that.equals(name);
            this.api.data.zones.length.should.equal(0);
          }).catch(done);
      });

      it('intial zones are available', function () {
        const index = this.api.data.zones.push({}) - 1;
        const zone = this.api.data.zones[index];
        return zone.loadId()
          .then(id => {
            this.api.data.zones.length.should.equal(1);
            id.should.equal(0);
            const api = new Api();
            return Promise.all([id, zone, api, api.loadData()]);
          })
          .then(([id, zone, api, data]) => {
            data.should.deep.equal(api.data);
            should.exist(data.zones.getById(id));
            data.zones.getById(id).should.deep.equal(zone);
          });
      });

      describe('points', () => {
        it('can push zone points', function () {
          const point = {x: 99, y: 12313};
          const index = this.api.data.zones.push({}) - 1;
          const zone = this.api.data.zones[index];
          return zone.loadId()
            .then(() => {
              zone.points.push(point);
              return zone.points[4].loadId();
            })
            .then(id => {
              id.should.equal(4);
              zone.points[4].should.deep.equal(point);
            })
            .then(() => {
              return zone.points._request.get('');
            })
            .then(json => {
              zone.points.toJSON().should.deep.equal(json);
              return Promise.all(zone.points.map(o => o.loadId()));
            })
            .then(ids => {
              ids.should.deep.equal([0, 1, 2, 3, 4]);
            });
        });

        it('can shift zone points', function () {
          const points = [{
            x: 1231,
            y: 132
          }, {
            x: 41583,
            y: 13224
          }, {
            x: 41122,
            y: 9998
          }, {
            x: 616,
            y: 2244
          }];
          const expected = Object.assign({}, points[0]);
          const index = this.api.data.zones.push({points: points.slice()}) - 1;
          const zone = this.api.data.zones[index];
          return zone.loadId()
            .then(() => {
              return new Promise(resolve => {
                let count = 0;
                const expected = 7;
                const removed = zone.points.shift();
                zone.on('gone', () => {
                  if (++count === expected) {
                    resolve(removed);
                  }
                });
                zone.on('committed', () => {
                  if (++count === expected) {
                    resolve(removed);
                  }
                });
              });
            })
            .then(shifted => {
              shifted.should.deep.equal(expected);
            })
            .then(() => {
              return zone.points._request.get('');
            })
            .then(json => {
              zone.points.toJSON().should.deep.equal(json);
              zone.points.toJSON().should.deep.equal(points.slice(1));
              return Promise.all(zone.points.map(o => o.loadId()));
            })
            .then(ids => {
              ids.should.deep.equal([0, 1, 2]);
            });
        });

        it('can splice zone points', function () {
          const points = [{
            x: 8828,
            y: 61789
          }, {
            x: 7828,
            y: 13143
          }, {
            x: 59879,
            y: 4978
          }, {
            x: 1578,
            y: 3070
          }];
          const added = [{
            x: 1689,
            y: 1552
          }, {
            x: 16789,
            y: 589
          }, {
            x: 9875,
            y: 62267
          }, {
            x: 6859,
            y: 11111
          }];
          const final = [points[0], ...added, points[3]];
          const expected = points.slice(1, 3).map(o => Object.assign({}, o));
          const index = this.api.data.zones.push({points: points.slice()}) - 1;
          const zone = this.api.data.zones[index];
          return zone.loadId()
            .then(() => {
              return new Promise(resolve => {
                let count = 0;
                const expected = 8;
                const removed = zone.points.splice(1, 2, ...added);
                zone.on('resolved', () => {
                  if (++count === expected) {
                    resolve(removed);
                  }
                });
                zone.on('committed', () => {
                  if (++count === expected) {
                    resolve(removed);
                  }
                });
              });
            })
            .then(removed => {
              removed.should.deep.equal(expected);
            })
            .then(() => {
              return zone.points._request.get('');
            })
            .then(json => {
              zone.points.toJSON().should.deep.equal(json);
              zone.points.toJSON().should.deep.equal(final);
              return Promise.all(zone.points.map(o => o.loadId()));
            })
            .then(ids => {
              ids.should.deep.equal([0, 1, 2, 3, 4, 5]);
            });
        });

        it('points have initial data', function () {
          const point = {x: 99, y: 12313};
          const index = this.api.data.zones.push({}) - 1;
          const zone = this.api.data.zones[index];
          return zone.loadId()
            .then(() => {
              zone.points.push(point);
              return zone.points[4].loadId();
            })
            .then(id => {
              id.should.equal(4);
              zone.points[4].should.deep.equal(point);
            })
            .then(() => {
              return zone.points._request.get('');
            })
            .then(json => {
              this.api = new Api();
              return Promise.all([json, this.api.loadData()]);
            })
            .then(([json, data]) => {
              data.zones[0].points.toJSON().should.deep.equal(json);
            });
        });
      });
    });

    describe('observables', () => {
      beforeEach(function () {
        this.api.data.observables.length.should.equal(0, 'the observables were not empty before the test started');
      });

      it('can push new observables cpu', function () {
        const index = this.api.data.observables.push({typename: 'cpu'}) - 1;
        const observable = this.api.data.observables[index];
        return observable.loadId()
          .then(id => {
            this.api.data.observables.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can push new observables test', function () {
        const index = this.api.data.observables.push({typename: 'test'}) - 1;
        const observable = this.api.data.observables[index];
        return observable.loadId()
          .then(id => {
            this.api.data.observables.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can add lots of observables', function () {
        this.api.data.observables.splice(0, 0, {typename: 'test'}, {typename: 'cpu', threshold: 0.6}, {typename: 'test', interval: 500}, {typename: 'cpu'});
        this.api.data.observables.length.should.equal(4);
        return Promise.all(this.api.data.observables.map(observables => observables.loadId()))
          .then(() => this.api._request.get('observables'))
          .then(observables => {
            Object.keys(observables).should.have.length(4);
          });
      });

      it('can unshift new observables', function () {
        this.api.data.observables.unshift({typename: 'cpu'});
        return this.api.data.observables[0].loadId()
          .then(id => {
            this.api.data.observables.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can pop observables', function (done) {
        const typename = 'cpu';
        const index = this.api.data.observables.push({typename}) - 1;
        const observables = this.api.data.observables[index];
        observables.loadId()
          .then(() => {
            this.api.data.observables.length.should.equal(1);
            this.api.data.observables.on('reinsert', done);
            this.api.data.observables.on('gone', () => done());
            this.api.data.observables.pop().should.have.property('typename').that.equals(typename);
          }).catch(done);
      });

      it('should error on adding observable with incorrect typename', function () {
        const fun = () => {
          this.api.data.observables.push({typename: 'blah'});
        };
        fun.should.throw(Error);
      });

      it('should error on adding observable with no data', function () {
        const fun = () => {
          this.api.data.observables.push({});
        };
        fun.should.throw(Error);
      });
    });

    describe('actions', () => {
      beforeEach(function () {
        this.api.data.actions.length.should.equal(0, 'the actions were not empty before the test started');
      });

      it('can push new actions tcp', function () {
        const index = this.api.data.actions.push({typename: 'tcp'}) - 1;
        const action = this.api.data.actions[index];
        return action.loadId()
          .then(id => {
            this.api.data.actions.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can add lots of actions', function () {
        this.api.data.actions.splice(0, 0, {typename: 'tcp'}, {typename: 'tcp', uri: 'test'}, {typename: 'tcp', template: 'test'}, {typename: 'tcp', uri: 'test1', template: 'test2'});
        this.api.data.actions.length.should.equal(4);
        return Promise.all(this.api.data.actions.map(actions => actions.loadId()))
          .then(() => this.api._request.get('actions'))
          .then(actions => {
            Object.keys(actions).should.have.length(4);
          });
      });

      it('can unshift new actions', function () {
        this.api.data.actions.unshift({typename: 'tcp'});
        return this.api.data.actions[0].loadId()
          .then(id => {
            this.api.data.actions.length.should.equal(1);
            return id;
          })
          .should.eventually.equal(0);
      });

      it('can pop actions', function (done) {
        const typename = 'tcp';
        const index = this.api.data.actions.push({typename}) - 1;
        const actions = this.api.data.actions[index];
        actions.loadId()
          .then(() => {
            this.api.data.actions.length.should.equal(1);
            this.api.data.actions.on('reinsert', done);
            this.api.data.actions.on('gone', () => done());
            this.api.data.actions.pop().should.have.property('typename').that.equals(typename);
          }).catch(done);
      });

      it('should error on adding actions with incorrect typename', function () {
        const fun = () => {
          this.api.data.actions.push({typename: 'blah'});
        };
        fun.should.throw(Error);
      });

      it('should error on adding actions with no data', function () {
        const fun = () => {
          this.api.data.actions.push({});
        };
        fun.should.throw(Error);
      });

      it('should be able to add an observe to an action', function () {
        const index = this.api.data.actions.push({typename: 'tcp'}) - 1;
        const action = this.api.data.actions[index];
        return action.loadId()
          .then(id => {
            this.api.data.actions.length.should.equal(1);
            id.should.equal(0);
            return id;
          })
          .then(() => {
            action.observe.push(0);
            const observable = action.observe._data[action.observe.length - 1];
            return observable.loadId()
              .then(id => {
                id.should.equal(0);
                return observable._request.get('');
              });
          })
          .should.eventually.equal(0);
      });
    });
  });
});
